@@ -179,7 +179,8 @@ let (s1,d1,w1,c1,re1,nyi1,t1) = statistic1 r 0 true in
 let (s2,_,w2,_,_,_,_) = statistic1 r 0 false in
 (s1,s2,d1,w1,w2,c1,re1,nyi1,t1)
 
-(** MAIN **)    
+(** MAIN **)
+let tstp : bool ref = ref false;;
 
 let print_coq_proofscript c r =
   
@@ -237,7 +238,14 @@ let print_coq_proofscript c r =
 	(*** Output Coq ***)
 
   let beforeCoq= Sys.time() in
-  if (result_coq && size < maxcoqproofsize ) then ref_coq c prenorm_refutation else raise (CoqProofTooBig size);
+  if (!tstp) then (*** output in tstp format ***)
+    begin
+      ref_tstp c prenorm_refutation
+    end
+  else
+    begin
+      if (result_coq && size < maxcoqproofsize ) then ref_coq c prenorm_refutation else raise (CoqProofTooBig size)
+    end;
   let coqTime= int_of_float ((Sys.time() -. beforeCoq) *. 1000.0) in
   if (result_coq ) then if (!verbosity > 3) then Printf.printf  "Coq output done: time %d  \n" coqTime ;
 	(*** Output Latex Search ***)
@@ -494,3 +502,9 @@ let print_hocore c r =
 	  ((ntmn,!conjecturename)::a)
       | _ -> a
     end
+
+(*** Chad - July 2012 - output proof script in tstp format. ***)
+let print_tstp c r =
+  tstp := true;
+  print_coq_proofscript c r
+
