@@ -24,9 +24,82 @@ let counterstrategy_schedule_1 = [
 ("satmode1",10.0);
 ];;
 
-(*** Satallax 2.5 schedule. ***)
-(*** 56 modes, 226.1 total time, 1999 problems ***)
-let strategy_schedule_1 = [
+(*** Satallax 2.7 schedule. ***)
+(*** 68 modes, 299.1 total time, 4865 problems ***)
+let strategy_schedule_2_7 = [
+("mode438",0.2);
+("mode439",0.2);
+("mode495",0.2);
+("mode371",2.7);
+("mode446",2.0);
+("mode447",2.8);
+("mode445",1.0);
+("mode456",0.3);
+("mode375",0.2);
+("mode405",0.6);
+("mode412",0.8);
+("mode455",1.9);
+("mode450",0.2);
+("mode496",2.6);
+("mode379",3.4);
+("mode448",0.2);
+("mode389",0.6);
+("mode388",3.7);
+("mode457",0.5);
+("mode453",0.2);
+("mode497",0.2);
+("mode454",1.7);
+("mode451",0.4);
+("mode400",0.3);
+("mode498",3.0);
+("mode459",0.4);
+("mode411",0.4);
+("mode461",0.2);
+("mode460",0.2);
+("mode393",0.2);
+("mode499",3.5);
+("mode458",1.2);
+("mode464",1.5);
+("mode485",2.3);
+("mode466",0.3);
+("mode467",0.3);
+("mode500",4.6);
+("mode398",3.1);
+("mode469",0.8);
+("mode471",0.9);
+("mode484",2.0);
+("mode473",0.5);
+("mode479",7.6);
+("mode501",7.0);
+("mode474",0.7);
+("mode502",2.2);
+("mode406",3.7);
+("mode213",2.4);
+("mode481",3.5);
+("mode493",13.6);
+("mode503",7.0);
+("mode504",1.4);
+("mode480",1.4);
+("mode483",5.8);
+("mode478",2.9);
+("mode482",1.8);
+("mode505",2.5);
+("mode490",11.8);
+("mode401",54.9);
+("mode487",6.2);
+("mode506",7.0);
+("mode507",3.5);
+("mode492",10.1);
+("mode368",25.8);
+("mode491",10.0);
+("mode508",11.8);
+("mode433",14.2);
+("mode494",28.0);
+];;
+
+(*** Satallax 2.5 schedule. Use this if a proof is requested. ***)
+(*** 56 modes, 226.1 total time ***)
+let strategy_schedule_2_5 = [
 ("mode175",0.1);
 ("mode288",0.1);
 ("mode238",0.1);
@@ -94,7 +167,9 @@ let print_help () =
    print_newline();
    print_string "-m : Mode";
    print_newline();
-   print_string "-P <file> : PicoMus binary file";
+   print_string "-P <file> : PicoMus executable file";
+   print_newline();
+   print_string "-E <file> : E prover executable file";
    print_newline();
    print_string "-V : Print version number and quit";
    print_newline();
@@ -152,7 +227,6 @@ let load_mode m =
   let modefile = (!modedir ^ "/" ^ m) in
   if (Sys.file_exists modefile) then
     begin
-      init_flags();
       try
 	read_mode_file (open_in modefile)
       with
@@ -237,7 +311,8 @@ let handle_slave_return pstatus =
     if (pstatus >= 10) then exit pstatus else ()
   end;;
 
-let problemfile = ref "";;
+let printnumaxiomsflag : bool ref = ref false;;
+let selectaxiomsflag : bool ref = ref false;;
 
 let process_command_line_args () =
   let argc = Array.length Sys.argv in
@@ -246,9 +321,26 @@ let process_command_line_args () =
     let i = ref 1 in
     while (!i < argc) do
       (match Sys.argv.(!i) with
-      | "-P" -> (incr i; slaveargs := ("-P"::!slaveargs); Config.picomus := Sys.argv.(!i))
+      | "-numaxioms" -> incr i; printnumaxiomsflag := true
+      | "-selectaxioms" -> (*** This is only to experiment with different selections and (order) of the axioms/conjecture. ***)
+	  begin (*** giving nothing for a problem with n axioms/conjecture is equivalent to giving -selectaxioms n (n-1) ... 0 ***)
+	    selectaxiomsflag := true;
+	    incr i; let num = (int_of_string Sys.argv.(!i)) in
+	    for j = 0 to num - 1 do
+	      incr i; let i2 = (int_of_string Sys.argv.(!i)) in
+	      select_axioms_list := i2 :: !select_axioms_list
+	    done
+	  end
+      | "-P" -> (incr i; slaveargs := (Sys.argv.(!i)::"-P"::!slaveargs); Config.picomus := Sys.argv.(!i))
+      | "-E" -> (incr i; slaveargs := (Sys.argv.(!i)::"-E"::!slaveargs); Config.eprover := Sys.argv.(!i))
       | "-m" -> (incr i; mode := ((Sys.argv.(!i)) :: (!mode)))
-      | "-M" -> (incr i; modedir := (Sys.argv.(!i)))
+      | "-M" ->
+	  begin
+	    incr i;
+	    let mdir = Sys.argv.(!i) in
+	    slaveargs := (mdir::"-M"::!slaveargs);
+	    modedir := mdir
+	  end
       | "-t" -> (incr i; timeout := (Some (float_of_string Sys.argv.(!i))))
       | "-slave" -> (incr i; slave := true)
       |	"-C" -> (incr i; slaveargs := ("-C"::!slaveargs); coqlocalfile := true)
@@ -300,6 +392,12 @@ let process_command_line_args () =
 	    slaveargs := ("-N"::!slaveargs);
 	    nontheorem := true;
 	  end
+      |	"-foffiles" -> (*** This is only for testing and debugging interaction with FO provers like E. ***)
+	  begin
+	    incr i;
+	    slaveargs := ("-foffiles"::!slaveargs);
+	    foffiles := true;
+	  end
       | "" -> raise (InputHelpError "Problem processing command line arguments")
       |	"-" -> incr i; problemfile := ""
       | option ->
@@ -317,12 +415,19 @@ let process_command_line_args () =
 	  incr i)
     done
   end
-    
+
+let check_search_pfterm_ok () =
+  if (get_bool_flag "USE_E") then
+    begin
+      raise (GenericError("Proofs cannot currently be reported when E is used (either set flag USE_E to false or do not use -p)"))
+    end;;
+
 let search_main () =
   match (!mode) with
   | (_::_) ->
       begin
 	try
+	  init_flags();
 	  ignore (List.map load_mode (!mode));
 	  begin
 	    basetypestobool := get_bool_flag "BASETYPESTOPROP";
@@ -355,6 +460,7 @@ let search_main () =
 	    | _ -> ()
 	  end;
 	  if (!verbosity > 8) then print_flags () else ();
+          (match (!mkproofterm) with Some _ -> check_search_pfterm_ok () | None -> ());
 	  let s = (get_timeout_default 0.0) in
 	  if (s > 0.0) then
 	    begin
@@ -362,6 +468,8 @@ let search_main () =
 	      if (!coq) then coq_init();
 	      if (!coqlocalfile) then read_coq_file (!problemfile) else read_thf_file (!problemfile) (find_read_thf_file (Filename.dirname (!problemfile)));
 	      if ((!coq) && (not (!coq2))) then print_coqsig !coqoutchannel;
+	      if (!printnumaxiomsflag) then print_num_axioms ();
+	      if (!selectaxiomsflag) then select_axioms ();
 	      search ()
 	    end
 	  else
@@ -369,6 +477,8 @@ let search_main () =
 	      if (!coq) then coq_init();
 	      if (!coqlocalfile) then read_coq_file (!problemfile) else read_thf_file (!problemfile) (find_read_thf_file (Filename.dirname (!problemfile)));
 	      if ((!coq) && (not (!coq2)) && (not (!slave))) then print_coqsig !coqoutchannel;
+	      if (!printnumaxiomsflag) then print_num_axioms ();
+	      if (!selectaxiomsflag) then select_axioms ();
 	      search ()
 	    end
 	with
@@ -464,10 +574,15 @@ let search_main () =
 	  begin
 	    counterstrategy_schedule_1
 	  end
-	else
+	else if (mkprooftermp ()) then
 	  begin
-	    strategy_schedule_1
-	  end in
+	    strategy_schedule_2_5
+	  end
+        else
+	  begin
+	    strategy_schedule_2_7
+	  end
+      in
       let scheduletime = List.fold_left (fun s1 (x,s2) -> (s1 +. s2)) 0.0 strategy_schedule in (*** total time of schedule ***)
       let timeoutfac = max 1. ((get_timeout_default scheduletime) /. scheduletime) in (*** timeout factor ***)
       let schedulenum = ref (List.length strategy_schedule) in
