@@ -13,6 +13,16 @@ exception FileNotFound of string
 exception GenericError of string
 exception CoqProofTooBig of int
 
+let useeprover = ref false;;
+let eeqoequiv = ref false;;
+let eproverperiod = ref 10;;
+let eprovertimeout = ref 1;;
+let foffiles = ref false;;
+
+type clause = int list
+let clauses : clause list ref = ref []
+let clausesTable : (clause,unit) Hashtbl.t = Hashtbl.create 10
+
 let slaveargs = ref [Sys.argv.(0)];;
 let mode : string list ref = ref []
 let timeout : float option ref = ref None
@@ -20,6 +30,7 @@ let verbosity : int ref = ref 1 (*** Default to 1; 0 should mean print nothing (
 let nontheorem : bool ref = ref false (*** March 2012 - Know if you're explicitly trying to determine Satisfiable/CounterSatisfiable ***)
 let coq = ref false
 let coq2 = ref false
+let problemfile = ref ""
 let coqlocalfile = ref false
 let coqglobalfile = ref false
 let coqinchannel : in_channel ref = ref stdin
@@ -48,8 +59,10 @@ let coqsig_hyp : (string * pretrm) list ref = ref []
 let coqsig_def_trm : (string * trm) list ref = ref []
 let coqsig_hyp_trm : (string * trm) list ref = ref []
 let name_base : (string,unit) Hashtbl.t = Hashtbl.create 10
+let name_base_list : string list ref = ref []
 let name_tp : (string,stp) Hashtbl.t = Hashtbl.create 100
 let name_trm : (string,(trm * stp) * bool ref) Hashtbl.t = Hashtbl.create 100
+let name_trm_list : (string * trm * stp) list ref = ref []
 let name_def : (string,trm) Hashtbl.t = Hashtbl.create 100
 let name_def_prenorm : (string,trm) Hashtbl.t = Hashtbl.create 100
 let name_hyp : (string,trm) Hashtbl.t = Hashtbl.create 100
@@ -67,6 +80,10 @@ let set_timer s =
   ignore (Sys.signal Sys.sigalrm (Sys.Signal_handle (fun signo -> raise Timeout)));
   ignore (Unix.setitimer Unix.ITIMER_REAL { Unix.it_interval = 0.0; Unix.it_value = s });;
 
+let remaining_time () =
+  let a = Unix.getitimer Unix.ITIMER_REAL in
+  a.Unix.it_value
+
 let mult_timeout f =
   match (!timeout) with
   | Some tm -> timeout := Some (tm *. f)
@@ -119,73 +136,10 @@ let rec get_fresh_name a =
     begin
       Hashtbl.add name_tp x a;
       Hashtbl.add name_trm x ((xa,a), ref false);
+      name_trm_list := (x,xa,a)::!name_trm_list;
       (x,xa)
     end
 
-let next_atom = ref 1
-let atom_hash : (trm,int) Hashtbl.t = Hashtbl.create 1000
-let atom_hash_rev : (int,trm) Hashtbl.t = Hashtbl.create 1000
-
-let max_atom () = (!next_atom) - 1
-
-let new_atom m =
-  let a = (!next_atom) in
-  Hashtbl.add atom_hash m a;
-  Hashtbl.add atom_hash_rev a m;
-  if (!verbosity > 4) then if (!verbosity > 8) then (Printf.printf "Atom %d: %s\n" a (trm_str m)) else (Printf.printf "Atom %d\n" a);
-  incr next_atom;
-  a
-
-let get_atom m =
-  try
-    Hashtbl.find atom_hash m
-  with
-  | Not_found -> new_atom m
-
-let get_literal m =
-  match (neg_body m) with
-    Some n ->
-      (- (get_atom n))
-  | None -> (get_atom m)
-let literal_to_trm l =
-  if (l < 0) then
-    (neg (Hashtbl.find atom_hash_rev (- l)))
-  else
-    (Hashtbl.find atom_hash_rev l)
-
-let initial_branch : trm list ref = ref []
-let initial_branch_prenorm : trm list ref = ref []
-
-type clause = int list
-
-let clauses : clause list ref = ref []
-
-let clausesTable : (clause,unit) Hashtbl.t = Hashtbl.create 10
-
-type searchoption =
- (*** ProcessProp1(m) For rules requiring one formula in the head (most rules) ***)
-    ProcessProp1 of trm
- (*** Mating(plit,nlit,pl,nl) where pl,nl and plit = lit (h pl) and nlit = lit ~(h nl) ***)
-  | Mating of int * int * trm list * trm list
- (*** Confront(nplit,nnlit,a,s,t,u,v) where s,t,u,v:a and nplit = lit (s = t) and nnlit is lit (u != v) ***)
-  | Confront of int * int * stp * trm * trm * trm * trm
- (*** DefaultElt(a) - create a default element of type a ***)
-  | DefaultElt of string
- (*** DefaultEltInst(a) - create a default element of type a ***)
-  | DefaultEltInst of string
- (*** NewInst(a,m) - put m:a in the set of instantiations ***)
-  | NewInst of stp * trm
- (*** EnumIterDeep(d,a) - Enumerate all terms (using the current constants) of depth d ***)
-  | EnumIterDeep of int * stp
- (*** EnumTp(d,ar,a) - work on enumerating types that can be used to choose a polymorphic head (Eq, Forall, Choice) ***)
-  | EnumTp of int * stp * stp
- (*** EnumAp(d,Gamma,sigmal,h,c) - ***)
-  | EnumAp of (int * stp list * stp list * trm * (trm -> unit))
- (*** Enum(d,Gamma,sigma,c) ***)
-  | Enum of (int * stp list * stp * (trm -> unit))
- (*** Filter - use Minisat to filter usable sets ***)
-  | Filter of int
-
 type ruleinfo =
  (*** NegPropRule(m) For rules requiring one negative formula in the head, except negated forall ***)
     NegPropRule of trm
@@ -217,6 +171,313 @@ type refut =
   | SearchR of clause list * (clause -> ruleinfo)
   | AssumptionConflictR
 
+exception Unsatisfiable of refut option
+exception Satisfiable
+
+let next_atom = ref 1
+let atom_hash : (trm,int) Hashtbl.t = Hashtbl.create 1000
+let atom_hash_rev : (int,trm) Hashtbl.t = Hashtbl.create 1000
+
+let max_atom () = (!next_atom) - 1
+
+let prop_fo_forms : (int * foform) list ref = ref []
+let stp_fo_forms : (stp,int * foform) Hashtbl.t = Hashtbl.create 10
+let fo_types_h : (stp,unit) Hashtbl.t = Hashtbl.create 10
+let fo_types : stp list ref = ref []
+let fo_ecounter : (stp,int ref) Hashtbl.t = Hashtbl.create 10
+
+let plit_fo_str a =
+  if a > 0 then
+    "p" ^ (string_of_int a)
+  else
+    "~ p" ^ (string_of_int (- a))
+
+let rec printf_fotrm s m =
+  match m with
+  | FOVar(x) -> Printf.fprintf s "%s" x
+  | FOFun(f,[]) -> Printf.fprintf s "c%s" f
+  | FOFun(f,(a::al)) ->
+      begin
+	Printf.fprintf s "f%s(" f;
+	printf_fotrm s a;
+	List.iter (fun t -> Printf.fprintf s ","; printf_fotrm s t) al;
+	Printf.fprintf s ")";
+      end
+
+let rec printf_fof s m =
+  match m with
+  | FOEq(t1,t2) ->
+      Printf.fprintf s "(";
+      printf_fotrm s t1;
+      Printf.fprintf s " = ";
+      printf_fotrm s t2;
+      Printf.fprintf s ")";
+  | FOPred(p,[]) ->
+      Printf.fprintf s "q%s" p;
+  | FOPred(p,(a::al)) ->
+      Printf.fprintf s "r%s(" p;
+      printf_fotrm s a;
+      List.iter (fun t -> Printf.fprintf s ","; printf_fotrm s t) al;
+      Printf.fprintf s ")";
+  | FOImp(m1,m2) ->
+      Printf.fprintf s "(";
+      printf_fof s m1;
+      Printf.fprintf s " => ";
+      printf_fof s m2;
+      Printf.fprintf s ")";
+  | FOEqu(m1,m2) ->
+      Printf.fprintf s "(";
+      printf_fof s m1;
+      Printf.fprintf s " <=> ";
+      printf_fof s m2;
+      Printf.fprintf s ")";
+  | FOAll(x,m1) ->
+      Printf.fprintf s "( ! [%s] : " x;
+      printf_fof s m1;
+      Printf.fprintf s ")";
+  | FOFal ->
+      Printf.fprintf s "$false"
+
+let printf_fof_file s b =
+  let axc = ref 0 in
+	   (*** Send all propositional clauses we've sent to MiniSat. This contains the propositional structure of the problem. ***)
+  List.iter (fun c ->
+    match c with
+    | (l1::ll) ->
+	incr axc;
+	Printf.fprintf s "fof(ax%d,axiom,( %s " (!axc) (plit_fo_str l1);
+	List.iter (fun l2 -> Printf.fprintf s " | %s" (plit_fo_str l2)) ll;
+	Printf.fprintf s ")).\n";
+    | [] -> () (*** should never happen ***)
+	  )
+    !clauses;
+	   (*** Relate Some propositional literals to first order formulas. ***)
+	   (*** First the ones that are really propositional. ***)
+  List.iter (fun (p,fof) ->
+    incr axc;
+    Printf.fprintf s "fof(ax%d,axiom,( %s <=> " (!axc) (plit_fo_str p);
+    printf_fof s fof;
+    Printf.fprintf s ")).\n"
+      )
+    (!prop_fo_forms);
+	   (*** Finally the ones that are specifically viewed as first order via the current type b. ***)
+  List.iter (fun (p,fof) ->
+    incr axc;
+    Printf.fprintf s "fof(ax%d,axiom,( %s <=> " (!axc) (plit_fo_str p);
+    printf_fof s fof;
+    Printf.fprintf s ")).\n"
+      )
+    (Hashtbl.find_all stp_fo_forms b)
+
+(*** Output FOF as Sexpr -- just for debugging BEGIN ***)
+let rec printf_fstrm s m =
+  match m with
+  | FOVar(x) -> Printf.fprintf s "|%s|" x
+  | FOFun(f,[]) -> Printf.fprintf s "|%s|" f
+  | FOFun(f,(a::al)) ->
+      begin
+	Printf.fprintf s "(|%s| " f;
+	printf_fstrm s a;
+	List.iter (fun t -> Printf.fprintf s " "; printf_fstrm s t) al;
+	Printf.fprintf s ")";
+      end
+
+let rec printf_fs s m =
+  match m with
+  | FOEq(t1,t2) ->
+      Printf.fprintf s "(= ";
+      printf_fstrm s t1;
+      Printf.fprintf s " ";
+      printf_fstrm s t2;
+      Printf.fprintf s ")";
+  | FOPred(p,[]) ->
+      Printf.fprintf s "|%s|" p;
+  | FOPred(p,(a::al)) ->
+      Printf.fprintf s "(|%s| " p;
+      printf_fstrm s a;
+      List.iter (fun t -> Printf.fprintf s " "; printf_fstrm s t) al;
+      Printf.fprintf s ")";
+  | FOImp(m1,m2) ->
+      Printf.fprintf s "(=> ";
+      printf_fs s m1;
+      Printf.fprintf s " ";
+      printf_fs s m2;
+      Printf.fprintf s ")";
+  | FOEqu(m1,m2) ->
+      Printf.fprintf s "(<=> ";
+      printf_fs s m1;
+      Printf.fprintf s " ";
+      printf_fs s m2;
+      Printf.fprintf s ")";
+  | FOAll(x,m1) ->
+      Printf.fprintf s "(ALL |%s| " x;
+      printf_fs s m1;
+      Printf.fprintf s ")";
+  | FOFal ->
+      Printf.fprintf s "FALSE"
+
+let printf_fs_file s b =
+  let axc = ref 0 in
+	   (*** Send all propositional clauses we've sent to MiniSat. This contains the propositional structure of the problem. ***)
+  List.iter (fun c ->
+    match c with
+    | (l1::ll) ->
+	incr axc;
+	Printf.fprintf s "(AXIOM %d (|%s|" (!axc) (plit_fo_str l1);
+	List.iter (fun l2 -> Printf.fprintf s " |%s|" (plit_fo_str l2)) ll;
+	Printf.fprintf s "))\n";
+    | [] -> () (*** should never happen ***)
+	  )
+    !clauses;
+	   (*** Relate Some propositional literals to first order formulas. ***)
+	   (*** First the ones that are really propositional. ***)
+  List.iter (fun (p,fof) ->
+    incr axc;
+    Printf.fprintf s "(EAXIOM %d |%s| " (!axc) (plit_fo_str p);
+    printf_fs s fof;
+    Printf.fprintf s ")\n"
+      )
+    (!prop_fo_forms);
+	   (*** Finally the ones that are specifically viewed as first order via the current type b. ***)
+  List.iter (fun (p,fof) ->
+    incr axc;
+    Printf.fprintf s "(EAXIOM %d |%s| " (!axc) (plit_fo_str p);
+    printf_fs s fof;
+    Printf.fprintf s ")\n"
+      )
+    (Hashtbl.find_all stp_fo_forms b)
+(*** Output FOF as Sexpr -- just for debugging END ***)
+
+(*** Remember if a is FO, so I can send this to E. Send it to E every E_PERIOD times. ***)
+let new_possible_fo_formula m a =
+ try
+   match trm_to_foform_stp m (!eeqoequiv) with
+   | (fof,Some b) ->
+       if (not (Hashtbl.mem fo_types_h b)) then
+	 begin
+	   if (!verbosity > 4) then (Printf.printf "New FO type: %s\n" (stp_str b); flush stdout);
+	   Hashtbl.add fo_types_h b ();
+	   fo_types := b::!fo_types;
+	   Hashtbl.add fo_ecounter b (ref 0);
+	 end;
+       if (!verbosity > 4) then (Printf.printf "Adding FOF for type %s: %s\n" (stp_str b) (trm_str m); flush stdout);
+       Hashtbl.add stp_fo_forms b (a,fof);
+       let r = Hashtbl.find fo_ecounter b in
+       incr r;
+       if (!r >= !eproverperiod) then
+	 begin
+	   r := 0;
+	   if (!verbosity > 4) then (Printf.printf "Calling E on FOF for type %s\n" (stp_str b); flush stdout);
+	   if (!foffiles) then
+	     begin
+	       let foffilename = "/tmp/fof" ^ (string_of_float (Unix.time ())) ^ "p" in
+	       let foffile = open_out foffilename in
+	       printf_fof_file foffile b;
+	       close_out foffile;
+	       Printf.printf "./E/PROVER/eprover -s --auto --cpu-limit=1 --tptp3-in < %s\n" foffilename;
+	       Unix.sleep 1; (*** sleep for a second so the names will be unique ***)
+	     end;
+(***
+ ***)
+	   let etimeoutnow =
+             match (!timeout) with
+             | Some _ -> min (int_of_float (remaining_time ())) (!eprovertimeout)
+             | None -> !eprovertimeout
+           in
+	   if (etimeoutnow > 0) then
+	     begin
+	       if (!verbosity > 4) then Printf.printf "Calling E for %d seconds.\n" etimeoutnow;
+	       let (myout,myin,myerr) = Unix.open_process_full ((!Config.eprover) ^ " -s --auto --cpu-limit=" ^ (string_of_int etimeoutnow) ^ "  --tptp3-in") [| |] in
+               printf_fof_file myin b;
+	       close_out myin;
+	       if (!verbosity > 4) then (Printf.printf "About to read result from eprover\n"; flush stdout);
+	       try
+		 while true do
+		   match (input_line myout) with
+		   | "# SZS status Unsatisfiable" ->
+                       if (!foffiles) then
+			 begin
+			   let foffilename = "/tmp/fof" ^ (string_of_float (Unix.time ())) ^ "p" in
+			   let foffile = open_out foffilename in
+			   printf_fof_file foffile b;
+			   close_out foffile;
+			   let fsfilename = "/tmp/fs" ^ (string_of_float (Unix.time ())) ^ ".lisp" in
+			   let fsfile = open_out fsfilename in
+			   printf_fs_file fsfile b;
+			   close_out fsfile;
+			   Printf.printf "./E/PROVER/eprover -s --auto --cpu-limit=1 --tptp3-in < %s\n" foffilename;
+			 end;
+                       ignore(Unix.close_process_full(myout,myin,myerr));
+		       raise (Unsatisfiable None)
+		   | _ -> ()
+		 done
+	       with End_of_file -> ignore(Unix.close_process_full(myout,myin,myerr))
+	     end
+	 end
+   | (fof,None) ->
+       if (!verbosity > 4) then (Printf.printf "Adding prop FOF: %s\n" (trm_str m); flush stdout);
+       prop_fo_forms := ((a,fof)::!prop_fo_forms)
+ with NotFO -> ()
+
+let new_atom m =
+  let a = (!next_atom) in
+  incr next_atom;
+  Hashtbl.add atom_hash m a;
+  Hashtbl.add atom_hash_rev a m;
+  if (!verbosity > 4) then if (!verbosity > 8) then (Printf.printf "Atom %d: %s\n" a (trm_str m)) else (Printf.printf "Atom %d\n" a);
+  if (!useeprover) then new_possible_fo_formula m a;
+  a
+
+let get_atom m =
+  try
+    Hashtbl.find atom_hash m
+  with
+  | Not_found -> new_atom m
+
+let get_literal m =
+  match (neg_body m) with
+    Some n ->
+      (- (get_atom n))
+  | None -> (get_atom m)
+
+let literal_to_trm l =
+  if (l < 0) then
+    (neg (Hashtbl.find atom_hash_rev (- l)))
+  else
+    (Hashtbl.find atom_hash_rev l)
+
+let assignedp m =
+  match (neg_body m) with
+    Some n -> Hashtbl.mem atom_hash n
+  | None -> Hashtbl.mem atom_hash m
+
+let initial_branch : trm list ref = ref []
+let initial_branch_prenorm : trm list ref = ref []
+
+type searchoption =
+ (*** ProcessProp1(m) For rules requiring one formula in the head (most rules) ***)
+    ProcessProp1 of trm
+ (*** Mating(plit,nlit,pl,nl) where pl,nl and plit = lit (h pl) and nlit = lit ~(h nl) ***)
+  | Mating of int * int * trm list * trm list
+ (*** Confront(nplit,nnlit,a,s,t,u,v) where s,t,u,v:a and nplit = lit (s = t) and nnlit is lit (u != v) ***)
+  | Confront of int * int * stp * trm * trm * trm * trm
+ (*** DefaultElt(a) - create a default element of type a ***)
+  | DefaultElt of string
+ (*** DefaultEltInst(a) - create a default element of type a ***)
+  | DefaultEltInst of string
+ (*** NewInst(a,m) - put m:a in the set of instantiations ***)
+  | NewInst of stp * trm
+ (*** EnumIterDeep(d,a) - Enumerate all terms (using the current constants) of depth d ***)
+  | EnumIterDeep of int * stp
+ (*** EnumTp(d,ar,a) - work on enumerating types that can be used to choose a polymorphic head (Eq, Forall, Choice) ***)
+  | EnumTp of int * stp * stp
+ (*** EnumAp(d,Gamma,sigmal,h,c) - ***)
+  | EnumAp of (int * stp list * stp list * trm * (trm -> unit))
+ (*** Enum(d,Gamma,sigma,c) ***)
+  | Enum of (int * stp list * stp * (trm -> unit))
+ (*** Filter - use Minisat to filter usable sets ***)
+  | Filter of int
+
 (*** For printing a searchoption - helpful for debugging. - Chad, Oct 2010 ***)
 let searchoption_str s =
   match s with
@@ -324,10 +585,14 @@ let insert_search_clause c r =
 
 (*** Different Implementations of Fair Priority Queues ***)
 module PQueue0 = PriorityQueue0 (struct type t = searchoption end)
-
 module PQueue1 = PriorityQueue1 (struct type t = searchoption end)
-
 module PQueue2 = PriorityQueue2 (struct type t = searchoption end)
+module PQueue3 = PriorityQueue3 (struct type t = searchoption end)
+module PQueue4 = PriorityQueue4 (struct type t = searchoption end)
+module PQueue5 = PriorityQueue5 (struct type t = searchoption end)
+module PQueue6 = PriorityQueue6 (struct type t = searchoption end)
+module PQueue7 = PriorityQueue7 (struct type t = searchoption end)
+module PQueue8 = PriorityQueue8 (struct type t = searchoption end)
 
 let insertWithPriority n o =
   let (ins,dp) =
@@ -336,6 +601,12 @@ let insertWithPriority n o =
       | 0 -> (PQueue0.insertWithPriority,PQueue0.debug_print)
       | 1 -> (PQueue1.insertWithPriority,PQueue1.debug_print)
       | 2 -> (PQueue2.insertWithPriority,PQueue2.debug_print)
+      |	3 -> (PQueue3.insertWithPriority,PQueue3.debug_print)
+      |	4 -> (PQueue4.insertWithPriority,PQueue4.debug_print)
+      |	5 -> (PQueue5.insertWithPriority,PQueue5.debug_print)
+      |	6 -> (PQueue6.insertWithPriority,PQueue6.debug_print)
+      |	7 -> (PQueue7.insertWithPriority,PQueue7.debug_print)
+      |	8 -> (PQueue8.insertWithPriority,PQueue8.debug_print)
       | n -> (raise (GenericError("Invalid value of PRIORITY_QUEUE_IMPL: " ^ (string_of_int n))))
     end
   in
@@ -352,6 +623,12 @@ let getNext () =
       | 0 -> (PQueue0.getNext,PQueue0.debug_print)
       | 1 -> (PQueue1.getNext,PQueue1.debug_print)
       | 2 -> (PQueue2.getNext,PQueue2.debug_print)
+      |	3 -> (PQueue3.getNext,PQueue3.debug_print)
+      |	4 -> (PQueue4.getNext,PQueue4.debug_print)
+      |	5 -> (PQueue5.getNext,PQueue5.debug_print)
+      |	6 -> (PQueue6.getNext,PQueue6.debug_print)
+      |	7 -> (PQueue7.getNext,PQueue7.debug_print)
+      |	8 -> (PQueue8.getNext,PQueue8.debug_print)
       | n -> (raise (GenericError("Invalid value of PRIORITY_QUEUE_IMPL: " ^ (string_of_int n))))
     end
   in
@@ -367,6 +644,12 @@ let peekAtNext() =
     | 0 -> PQueue0.peekAtNext()
     | 1 -> PQueue1.peekAtNext()
     | 2 -> PQueue2.peekAtNext()
+    | 3 -> PQueue3.peekAtNext()
+    | 4 -> PQueue4.peekAtNext()
+    | 5 -> PQueue5.peekAtNext()
+    | 6 -> PQueue6.peekAtNext()
+    | 7 -> PQueue7.peekAtNext()
+    | 8 -> PQueue8.peekAtNext()
     | n -> (raise (GenericError("Invalid value of PRIORITY_QUEUE_IMPL: " ^ (string_of_int n))))
   end
 
@@ -402,6 +685,7 @@ let neqns : (string,int * trm * trm) Hashtbl.t = Hashtbl.create 10
 
 let univpreds : (stp,(int * trm)) Hashtbl.t = Hashtbl.create 10
 let instantiations : (stp,(trm,unit) Hashtbl.t) Hashtbl.t = Hashtbl.create 10
+let instantiationslist : (stp,trm list) Hashtbl.t = Hashtbl.create 10
 let default_elts : (string,trm) Hashtbl.t = Hashtbl.create 10
 
 let set_default_elt aname x =
@@ -430,10 +714,7 @@ let default_elt_p aname =
 
 let get_instantiations a =
   try
-    let h = Hashtbl.find instantiations a in
-    let l = ref [] in
-    Hashtbl.iter (fun key _ -> l := (key::!l)) h;
-    !l
+    Hashtbl.find instantiationslist a
   with Not_found -> []
 
 let known_instantiation a m =
@@ -443,15 +724,43 @@ let known_instantiation a m =
     true
   with Not_found -> false
 
+let cons_instantiation m ml =
+    let iordcyc = get_int_flag "INSTANTIATION_ORDER_CYCLE" in
+    let iordmask = get_int_flag "INSTANTIATION_ORDER_MASK" in
+    if (iordcyc < 2) then
+      begin
+	if (iordmask mod 2 = 0) then
+	  (m::ml)
+	else
+	  (ml @ [m])
+      end
+    else
+      let j = List.length ml mod iordcyc in
+      begin
+	if ((iordmask lsr j) mod 2 = 0) then
+	  (m::ml)
+	else
+	  (ml @ [m])
+      end
+
+let add_instantiation_2 a m =
+  try
+    let ml = Hashtbl.find instantiationslist a in
+    Hashtbl.replace instantiationslist a (cons_instantiation m ml)
+  with Not_found ->
+    Hashtbl.add instantiationslist a [m]
+
 let add_instantiation a m =
   try
     let h = Hashtbl.find instantiations a in
     Hashtbl.add h m ();
+    add_instantiation_2 a m;
     set_default_elt_if_needed a m
   with Not_found ->
     let h : (trm,unit) Hashtbl.t = Hashtbl.create 10 in
     Hashtbl.add instantiations a h;
     Hashtbl.add h m ();
+    add_instantiation_2 a m;
     set_default_elt_if_needed a m
 
 let choiceopnames : (string,(stp * trm)) Hashtbl.t = Hashtbl.create 10
@@ -524,9 +833,6 @@ let decomposable x =
    | _ -> true
    with Not_found -> true (*** A name is decomposable by default ***)
 
-exception Unsatisfiable of refut option
-exception Satisfiable
-
 (*** Set completep to false if I use a mode that makes the search incomplete, so that failure does not imply satisfiability.
      Currently there is no such mode. - Chad, Oct 2010 ***)
 let completep = ref true
@@ -714,6 +1020,7 @@ let declare_base_type a =
       coqsig_base := (y::!coqsig_base)
     end;
   Hashtbl.add name_base a ();
+  name_base_list := (a::!name_base_list);
   if (get_bool_flag "CHOICE_AS_DEFAULT") then
     set_default_elt a (norm name_def (ap(Choice (Base a),Lam(Base a,False))))
 
@@ -755,7 +1062,7 @@ let st_to_trm m =
 
 let declare_typed_constant (name:string) (role:string) (m:pretrm) =
   begin
-    if (!verbosity > 4) then Printf.printf "declare_typed_constant %s %s\n%s\n" name role (pretrm_str m);
+    if (!verbosity > 4) then (Printf.printf "declare_typed_constant %s %s\n%s\n" name role (pretrm_str m); flush stdout);
     match m with
       POf(PName(x),m) ->
 	begin
@@ -778,7 +1085,8 @@ let declare_typed_constant (name:string) (role:string) (m:pretrm) =
 		  coqsig_const := (y,tp)::!coqsig_const
 		end;
 	      Hashtbl.add name_tp x tp;
-	      Hashtbl.add name_trm x ((Name(x,tp),tp),ref false)
+	      Hashtbl.add name_trm x ((Name(x,tp),tp),ref false);
+	      name_trm_list := (x,Name(x,tp),tp)::!name_trm_list
 	end
     | _ -> raise (GenericError ("Incorrect format for type declaration " ^ name))
   end
@@ -839,13 +1147,14 @@ let declare_definition_real x m =
 	  end;
 	Hashtbl.add name_tp x tp;
 	Hashtbl.add name_trm x ((Name(x,tp),tp),ref false);
+	name_trm_list := (x,Name(x,tp),tp)::!name_trm_list;
 	Hashtbl.add name_def_prenorm x tm;
 	Hashtbl.add name_def x (norm name_def (logicnorm tm))
       end
 
 let rec declare_thf_logic_formula (name:string) (role:string) (m:pretrm) =
   begin
-    if !verbosity > 4 then Printf.printf "declare_thf_logic_formula %s %s\n" name role;
+    if !verbosity > 4 then (Printf.printf "declare_thf_logic_formula %s %s\n" name role; flush stdout);
     if ((role = "axiom") || (role = "hypothesis") || (role = "lemma")) then
       begin
 	if (Hashtbl.mem name_hyp name) then raise (Redeclaration name);
@@ -887,7 +1196,7 @@ and declare_definition (name:string) (role:string) (m:pretrm) =
   try
     begin
       if (get_bool_flag "ALL_DEFS_AS_EQNS") then raise TreatAsAssumption;
-      if !verbosity > 4 then Printf.printf "declare_definition %s %s\n" name role;
+      if !verbosity > 4 then (Printf.printf "declare_definition %s %s\n" name role; flush stdout);
       match m with
 	PDef(PName(x),m) -> (*** No longer THF syntax. ***)
 	  declare_definition_real x m
@@ -977,6 +1286,11 @@ let reset_search () =
     Hashtbl.clear usableHeads_rtp;
     Hashtbl.clear choiceopnames;
     Hashtbl.clear filtered;
+    prop_fo_forms := [];
+    Hashtbl.clear stp_fo_forms;
+    Hashtbl.clear fo_types_h;
+    fo_types := [];
+    Hashtbl.clear fo_ecounter;
     search_init()
   end
 
@@ -994,3 +1308,27 @@ let print_model b =
 	  Printf.printf "%d%s %s\n" i (if (v = 0) then "+" else "-") (trm_str m)
       )
     atom_hash
+
+let rec select_list_rec l sl rl =
+  match sl with
+  | (j::sr) -> select_list_rec l sr (List.nth l j::rl)
+  | [] -> rl
+
+let select_axioms_list : int list ref = ref [];;
+
+let select_axioms () =
+  (*** I should select the prenorm branch too, but that's only needed for proof reconstruction. ***)
+  initial_branch := select_list_rec (!initial_branch) (!select_axioms_list) [];
+  if (!verbosity > 4) then
+    let i = ref 0 in
+    begin
+      Printf.printf "Initial branch after selection:\n";
+      incr i;
+      List.iter (fun m -> Printf.printf "%d %s\n" (!i) (trm_str m)) (!initial_branch)
+    end;;
+
+let print_num_axioms () =
+  let n = List.length (!initial_branch) in
+  Printf.printf "(NUMAXIOMS \"%s\" %d)\n" (!problemfile) n;
+  exit 123
+  
