@@ -109,3 +109,18 @@ val head_spine: trm -> trm * trm list
 val rtp : stp -> stp
 val argtps_rtp : stp -> stp list * stp
 
+(*** First Order Formulas - Dec 2012 ***)
+type fotrm =
+    FOVar of string
+  | FOFun of string * fotrm list
+type foform = (*** assuming a single sort ***)
+    FOEq of fotrm * fotrm
+  | FOPred of string * fotrm list
+  | FOImp of foform * foform
+  | FOEqu of foform * foform
+  | FOAll of string * foform
+  | FOFal
+
+exception NotFO
+
+val trm_to_foform_stp : trm -> bool -> foform * stp option
